/*!\file:  SpawnCoreParallel.cpp
 * \brief: run core ISSM solution using Dakota inputs coming from CPU 0.
 * \sa qmu.cpp DakotaPlugin.cpp
 *
 * This routine needs to be understood simultaneously with qmu.cpp and DakotaPlugin. 
 * SpawnCoreParallel is called by all CPUS, with CPU 0 holding Dakota variable values, along 
 * with variable descriptors. 
 *
 * SpawnCoreParallel takes care of broadcasting the variables and their descriptors across the MPI 
 * ring. Once this is done, we use the variables to modify the inputs for the solution core. 
 * For ex, if "rho_ice" is provided, for ex 920, we include "rho_ice" in the inputs, then 
 * call the core with the modified inputs. This is the way we get Dakota to explore the parameter 
 * spce of the core. 
 *
 * Once the core is called, we process the results of the core, and using the processed results, 
 * we compute response functions. The responses are computed on all CPUS, but they are targeted 
 * for CPU 0, which will get these values back to the Dakota engine. 
 *
 */ 

#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#error "Cannot compile with HAVE_CONFIG_H symbol! run configure first!"
#endif

#undef __FUNCT__ 
#define __FUNCT__ "SpawnCoreParallel"

#include "../objects/objects.h"
#include "../io/io.h"
#include "../EnumDefinitions/EnumDefinitions.h"
#include "../shared/shared.h"
#include "./Qmux.h"
#include "../include/macros.h"
#include "../parallel/parallel.h"

void SpawnCoreParallel(double* responses, int numresponses, double* variables, char** variables_descriptors,int numvariables, Model* model,ParameterInputs* inputs,int analysis_type,int sub_analysis_type,int counter){

	int i;
	
	/*output from core solutions: */
	DataSet* results=NULL;

	char** responses_descriptors=NULL;
	Param* param=NULL;
	char*  string=NULL;
	int    string_length;
	double* qmu_part=NULL;
	int     qmu_npart;
	int     debug=0;

	extern int my_rank;
	
	
	/*synchronize all cpus, as CPU 0 is probably late (it is starting the entire dakota strategy!) : */
	MPI_Barrier(MPI_COMM_WORLD);
	
	/*some parameters needed: */
	model->FindParam(&debug,"debug");
		
	/*First off, recover the response descriptors for the response functions: */
	param=(Param*)model->GetFormulation(DiagnosticAnalysisEnum(),HorizAnalysisEnum())->parameters->FindParamObject("responsedescriptors");
	if(!param)throw ErrorException(__FUNCT__," could not find response descriptors!");

	param->GetParameterValue(&responses_descriptors);

	/*Recover partitioning for dakota: */
	model->FindParam(&qmu_npart,"qmu_npart");
	model->FindParam(&qmu_part,"qmu_part");
	#ifdef _DEBUG_
	for(i=0;i<numresponses;i++){
		PetscSynchronizedPrintf(MPI_COMM_WORLD,"response descriptor %i: %s\n",i,responses_descriptors[i]);
		PetscSynchronizedFlush(MPI_COMM_WORLD);
	}
	#endif

	/*broadcast variables: only cpu 0 has correct values*/
	MPI_Bcast(&numvariables,1,MPI_INT,0,MPI_COMM_WORLD); 
	if(my_rank!=0)variables=(double*)xmalloc(numvariables*sizeof(double));
	MPI_Bcast(variables,numvariables,MPI_DOUBLE,0,MPI_COMM_WORLD); 

	#ifdef _DEBUG_
	for(i=0;i<numvariables;i++){
		PetscSynchronizedPrintf(MPI_COMM_WORLD,"variable %i: %g\n",i,variables[i]);
		PetscSynchronizedFlush(MPI_COMM_WORLD);
	}
	#endif

	/*broadcast variables_descriptors: */
	if(my_rank!=0){
		variables_descriptors=(char**)xmalloc(numvariables*sizeof(char*));
	}
	for(i=0;i<numvariables;i++){
		if(my_rank==0){
			string=variables_descriptors[i];
			string_length=(strlen(string)+1)*sizeof(char);
		}
		MPI_Bcast(&string_length,1,MPI_INT,0,MPI_COMM_WORLD); 
		if(my_rank!=0)string=(char*)xmalloc(string_length);
		MPI_Bcast(string,string_length,MPI_CHAR,0,MPI_COMM_WORLD); 
		if(my_rank!=0)variables_descriptors[i]=string;
	}

	#ifdef _DEBUG_
	for(i=0;i<numvariables;i++){
		PetscSynchronizedPrintf(MPI_COMM_WORLD,"variable descriptor %i: %s value: %g\n",i,variables_descriptors[i],variables[i]);
		PetscSynchronizedFlush(MPI_COMM_WORLD);
	}
	#endif

	_printf_("qmu iteration: %i\n",counter);

	//initialize results:
	results=new DataSet(ResultsEnum());

	/*Modify core inputs to reflect the dakota variables inputs: */
	inputs->UpdateFromDakota(variables,variables_descriptors,numvariables,model->GetFormulation(DiagnosticAnalysisEnum(),HorizAnalysisEnum())->parameters,qmu_part,qmu_npart); //diagnostic horiz model is the one holding the parameters for Dakota.

	/*Run the analysis core solution sequence, with the updated inputs: */
	if(analysis_type==DiagnosticAnalysisEnum()){
			
		if(debug)_printf_("Starting diagnostic core\n");

		diagnostic_core(results,model,inputs);

	}
	else if(analysis_type==ThermalAnalysisEnum()){
		
		if(debug)_printf_("Starting thermal core\n");
		thermal_core(results,model,inputs);

	}
	else if(analysis_type==PrognosticAnalysisEnum()){

		if(debug)_printf_("Starting prognostic core\n");
		prognostic_core(results,model,inputs);

	}
	else if(analysis_type==TransientAnalysisEnum()){

		if(debug)_printf_("Starting transient core\n");
		transient_core(results,model,inputs);

	}
	else throw ErrorException(__FUNCT__,exprintf("%s%i%s%i%s"," analysis_type ",analysis_type," and sub_analysis_type ",sub_analysis_type," not supported yet!"));
	
		

	/*Now process the outputs, before computing the dakota responses: */
	if(debug)_printf_("process results:\n");
	ProcessResults(&results,model,analysis_type); 
	

	/*compute responses on cpu 0: dummy for now! */
	if(debug)_printf_("compute dakota responses:\n");
	DakotaResponses(responses,responses_descriptors,numresponses,model,results,analysis_type,sub_analysis_type);

	/*Free ressources:*/
	delete results;

	//variables only on cpu != 0
	if(my_rank!=0){
		xfree((void**)&variables);
		for(i=0;i<numvariables;i++){
			string=variables_descriptors[i];
			xfree((void**)&string);
		}
		xfree((void**)&variables_descriptors);
	}
	//responses descriptors
	for(i=0;i<numresponses;i++){
		string=responses_descriptors[i];
		xfree((void**)&string);
	}
	//rest of dynamic allocations.
	xfree((void**)&responses_descriptors);
	xfree((void**)&qmu_part);
}

