/*!\file Model.c
 * \brief: implementation of the Model object
 */

#include "./objects.h"
#include "../io/io.h"
#include "../shared/shared.h"
#include "../include/macros.h"
#include "../issm.h"

#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#error "Cannot compile with HAVE_CONFIG_H symbol! run configure first!"
#endif

#include "./Model.h"
#include "stdio.h"

Model::Model(){

	femmodels=new DataSet(0);
	active=NULL;

}

Model::~Model(){

	delete femmodels;
	/*do not delete active, it just points to one of the femmodels!*/
}

#undef __FUNCT__
#define __FUNCT__ "Model::Echo"

void Model::Echo(void){

	printf("Models echo: \n");
	printf("   number of fem models: %i\n",femmodels->Size());

}
#undef __FUNCT__
#define __FUNCT__ "Model::DeepEcho"

void Model::DeepEcho(void){

	femmodels->Echo();

}
		
#undef __FUNCT__
#define __FUNCT__ "Model::AddFormulation"
void  Model::AddFormulation(ConstDataHandle MODEL, int analysis_type,int sub_analysis_type){

	/*FemModel: */
	FemModel*           femmodel=NULL;
	DataSet*            elements=NULL;
	DataSet*            nodes=NULL;
	DataSet*            constraints=NULL;
	DataSet*            loads=NULL;
	DataSet*            materials=NULL;
	DataSet*            parameters=NULL;
	Vec                 partition=NULL;
	Vec                 tpartition=NULL;
	Vec                 yg=NULL;
	Mat                 Rmg=NULL;
	Mat                 Gmn=NULL;
	NodeSets*           nodesets=NULL;
	Vec                 ys=NULL;
	Vec                 ys0=NULL;

	/*intermediary: */
	IoModel* iomodel=NULL;
	
	_printf_("   fill model with matlab workspace data\n");
	IoModelInit(&iomodel,MODEL); 

	_printf_("   specifying analysis\n");
	if (analysis_type!=0){
		iomodel->analysis_type=analysis_type;
	}
	if (sub_analysis_type!=0){
		iomodel->sub_analysis_type=sub_analysis_type;
	}

	_printf_("   create datasets:\n");
	CreateDataSets(&elements,&nodes,&materials,&constraints,&loads,&parameters,iomodel,MODEL);

	_printf_("   create degrees of freedom: \n");
	Dofx( &partition,&tpartition,elements,nodes, parameters);
	
	_printf_("   create single point constraints: \n");
	SpcNodesx( &yg, nodes,constraints); 
	
	_printf_("   create rigid body constraints:\n");
	MpcNodesx( &Rmg, nodes,constraints); 
	
	_printf_("   create node sets:\n");
	BuildNodeSetsx(&nodesets, nodes);

	_printf_("   reducing single point constraints vector:\n");
	Reducevectorgtosx(&ys,&ys0, yg,nodesets);
	
	_printf_("   normalizing rigid body constraints matrix:\n");
	NormalizeConstraintsx(&Gmn, Rmg,nodesets);

	_printf_("   configuring element and loads:\n");
	ConfigureObjectsx(elements, loads, nodes, materials);

	_printf_("   process parameters:\n");
	ProcessParamsx( parameters, partition);

	_printf_("   free ressources:\n");
	DeleteIoModel(&iomodel);

	/*Use all the data created to create an femmodel: */
	femmodel=new FemModel(elements,nodes,constraints,loads,materials,parameters,
			              partition,tpartition,yg,Rmg,Gmn,nodesets,ys,ys0);

	/*Add to femmodels in model: */
	femmodels->AddObject((Object*)femmodel);

}


#undef __FUNCT__
#define __FUNCT__ "Model::FindParam"

int   Model::FindParam(int* pparameter,char* parametername){

	FemModel* femmodel=NULL;

	/*no analysis_type or sub_analysis_type, find the parameter in the first dataset we find: */
	if (!femmodels->Size())throw ErrorException(__FUNCT__," no fem models were found!");

	/*recover first femmodel: */
	femmodel=(FemModel*)femmodels->GetObjectByOffset(0);

	if(!femmodel)return 0;

	femmodel->FindParam((void*)pparameter,parametername);

}
int   Model::FindParam(double* pparameter,char* parametername){
	
	FemModel* femmodel=NULL;

	/*no analysis_type or sub_analysis_type, find the parameter in the first dataset we find: */
	if (!femmodels->Size())throw ErrorException(__FUNCT__," no fem models were found!");

	/*recover first femmodel: */
	femmodel=(FemModel*)femmodels->GetObjectByOffset(0);
	
	if(!femmodel)return 0;
	
	femmodel->FindParam((void*)pparameter,parametername);


}
int   Model::FindParam(double** pparameter,char* parametername){
	
	FemModel* femmodel=NULL;

	/*no analysis_type or sub_analysis_type, find the parameter in the first dataset we find: */
	if (!femmodels->Size())throw ErrorException(__FUNCT__," no fem models were found!");

	/*recover first femmodel: */
	femmodel=(FemModel*)femmodels->GetObjectByOffset(0);
	
	if(!femmodel)return 0;
	
	femmodel->FindParam((void*)pparameter,parametername);


}
int   Model::FindParam(char** pparameter,char* parametername){
	
	FemModel* femmodel=NULL;

	/*no analysis_type or sub_analysis_type, find the parameter in the first dataset we find: */
	if (!femmodels->Size())throw ErrorException(__FUNCT__," no fem models were found!");

	/*recover first femmodel: */
	femmodel=(FemModel*)femmodels->GetObjectByOffset(0);
	
	if(!femmodel)return 0;
	
	femmodel->FindParam((void*)pparameter,parametername);

}

int   Model::FindParam(int* pparameter,char* parametername,int analysis_type,int sub_analysis_type){
	
	FemModel* femmodel=NULL;

	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type,sub_analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}

int   Model::FindParam(double* pparameter,char* parametername,int analysis_type,int sub_analysis_type){
	
	FemModel* femmodel=NULL;

	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type,sub_analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}


int   Model::FindParam(double** pparameter,char* parametername,int analysis_type,int sub_analysis_type){

	FemModel* femmodel=NULL;

	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type,sub_analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}

int   Model::FindParam(char** pparameter,char* parametername,int analysis_type,int sub_analysis_type){

	FemModel* femmodel=NULL;
	
	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type,sub_analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}

int   Model::FindParam(int* pparameter,char* parametername,int analysis_type){
	
	FemModel* femmodel=NULL;

	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}

int   Model::FindParam(double* pparameter,char* parametername,int analysis_type){
	
	FemModel* femmodel=NULL;

	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}


int   Model::FindParam(double** pparameter,char* parametername,int analysis_type){

	FemModel* femmodel=NULL;

	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}

int   Model::FindParam(char** pparameter,char* parametername,int analysis_type){

	FemModel* femmodel=NULL;
	
	/*find the correct formulation: */
	femmodel=this->GetFormulation(analysis_type);
	
	if(!femmodel)return 0;

	/*extract our parameter from the found formulation: */
	femmodel->FindParam((void*)pparameter,parametername);
}	

#undef __FUNCT__
#define __FUNCT__ "Model::GetFormulation"

FemModel* Model::GetFormulation(int analysis_type,int sub_analysis_type){
	
	int i;
	FemModel* femmodel=NULL;
	int femmodel_analysis_type;
	int femmodel_sub_analysis_type;
	int found=0;
	
	if (!femmodels->Size())throw ErrorException(__FUNCT__," no fem models were found!");

	/*find femmodel with correct analysis_type and sub_analysis_type: */
	for(i=0;i<femmodels->Size();i++){

		femmodel=(FemModel*)femmodels->GetObjectByOffset(i);
		femmodel->FindParam((void*)&femmodel_analysis_type,"analysis_type");
		femmodel->FindParam((void*)&femmodel_sub_analysis_type,"sub_analysis_type");

		if((analysis_type==femmodel_analysis_type) && (sub_analysis_type==femmodel_sub_analysis_type)){
			found=1;
			break;
		}
	}

	if(!found)return NULL;
	else return femmodel;
}
		
#undef __FUNCT__
#define __FUNCT__ "Model::GetFormulation"

FemModel* Model::GetFormulation(int analysis_type){
	
	int i;
	FemModel* femmodel=NULL;
	int femmodel_analysis_type;
	int femmodel_sub_analysis_type;
	int found=0;
	
	if (!femmodels->Size())throw ErrorException(__FUNCT__," no fem models were found!");

	/*find femmodel with correct analysis_type and sub_analysis_type: */
	for(i=0;i<femmodels->Size();i++){

		femmodel=(FemModel*)femmodels->GetObjectByOffset(i);
		femmodel->FindParam((void*)&femmodel_analysis_type,"analysis_type");

		if((analysis_type==femmodel_analysis_type)){
			found=1;
			break;
		}
	}

	if(!found)return NULL;
	else return femmodel;
}
		

FemModel* Model::GetActiveFormulation(){return active;}
		

void* Model::SetActiveFormulation(FemModel* femmodel){
	active=femmodel;
}

